/** 
 * @file iterator.cs 
 * Classes to implement value iterators for several data types. They are used
 * to loop over all possible method calls, method return types, stub property
 * values, etc. They implement the standard IEnumerator interface.
 *
 * (C) 2004 Martin Pitt <martin@piware.de>
 *
 * This software is distributed under the terms and conditions of the GNU
 * General Public License (see file COPYING)
 */

using System;
using System.Collections;

namespace CCM {

/// Iterator that loops over false and true.
public class BoolIterator : IEnumerator {
    public BoolIterator() { reset = true; }
    public void Reset() { reset = true; }

    public bool MoveNext() {
	if( reset ) {
	    v = false;
	    reset = false;
	    return true;
	}
	if( v ) {
	    reset = true;
	    return false;
	}
	v = true;
	return true;
    }

    public Object Current { 
	get { return v; }
    }

    public override string ToString() {
	return "({false,true}, " + v + ")";
    }

    private bool v;
    private bool reset;
}

/// Iterator that loops over a range of integers.
public class IntIterator : IEnumerator {
    /// Create iterator with the given boundaries and step.
    public IntIterator( int start, int end, int step ) {
	this.start = start;
	this.end = end;
	this.step = step;
	reset = true;
    }

    /**
     * Create iterator from a Range(min,max[,step]) attribute. If attr is
     * null, it defaults to IntIterator( -1, 1, 1 ). If step is not given, it
     * defaults to 1.
     */
    public IntIterator( Attribute attr ) {
	reset = true;
	if( attr == null ) {
	    start = -1;
	    end = 1;
	    step = 1;
	    return;
	}

	if( attr.positional.Count >= 2 ) {
	    start = getBound( attr.positional[0] );
	    end = getBound( attr.positional[1] );
	    step = (attr.positional.Count >= 3) ? 
		getBound( attr.positional[2] ) : 1;
	} else
	    throw new Exception( "IntIterator ctor( Attribute ): " + 
		    this + ": invalid Range attribute " + attr );

	if( start > end )
	    throw new Exception( "IntIterator ctor( Attribute ): " + 
		    this + ": invalid range in Range attribute " + attr );
    }

    /**
     * Return integer boundary described by b. If b is an int, returns it; if
     * b is a string, it is interpreted as class field and its value is
     * returned.
     */
    private static int getBound( Object b ) {
	if( b is int )
	    return (int) b;
	if( b is string ) {
	    Variable f = Class.currentClass().getField( (string) b );
	    if( f != null && f.type == 0.GetType() )
		return (int) f.value;
	    else
		throw new Exception( "IntIterator.getBound(): could not determine dynamic bound: " + b + 
			" is no integer field" );
	}

	throw new Exception( "IntIterator.getBound(): illegal boundary " +
		b.GetType() + " " + b );
    }

    public void Reset() { reset = true; }

    public bool MoveNext() {
	if( reset ) {
	    reset = false;
	    v = start;
	    return true;
	}
	v += step;
	if( v > end ) {
	    reset = true;
	    return false;
	}
	return true;
    }

    public Object Current { 
	get { return v; }
    }

    public override string ToString() {
	return "([" + start+ ", " + end + "]/" + step + ", " + v + ")";
    }

    private int start, end, step, v;
    private bool reset;
}

/// Iterator that loops over a range of doubles.
public class DoubleIterator : IEnumerator {
    /// Create iterator with the given boundaries and step.
    public DoubleIterator( double start, double end, double step ) {
	this.start = this.v = start;
	this.end = end;
	this.step = step;
	reset = true;
    }

    /**
     * Create iterator from a Range(min,max[,step]) attribute. If attr is
     * null, it defaults to IntIterator( -1.0, 1.0, 0.5 ). If step is not
     * given, it defaults to 0.5.
     */
    public DoubleIterator( Attribute attr ) {
	reset = true;
	if( attr == null ) {
	    start = -1.0;
	    end = 1.0;
	    step = 0.5;
	    return;
	}

	if( attr.positional.Count >= 2 ) {
	    start = getBound( attr.positional[0] );
	    end = getBound( attr.positional[1] );
	    step = (attr.positional.Count >= 3) ? 
		getBound( attr.positional[2] ) : 0.5;
	} else
	    throw new Exception( "DoubleIterator ctor( Attribute ): " + 
		    attr + ": invalid Range attribute" );
    }

    /**
     * Return double boundary described by b. If b is a double, returns it; if
     * b is a string, it is interpreted as class field and its value is
     * returned.
     */
    private static double getBound( Object b ) {
	if( b is double )
	    return (double) b;
	if( b is string ) {
	    Variable f = Class.currentClass().getField( (string) b );
	    if( f != null && f.type == (0.0d).GetType() )
		return (double) f.value;
	    else
		throw new Exception( "DoubleIterator.getBound(): could not determine dynamic bound: " + 
			b + " is no double field" );
	}

	throw new Exception( "DoubleIterator.getBound(): illegal boundary " 
		+ b.GetType() + " " + b );
    }

    public void Reset() { reset = true; }

    public bool MoveNext() {
	if( reset ) {
	    reset = false;
	    v = start;
	    return true;
	}
	v += step;
	if( v > end ) {
	    reset = true;
	    return false;
	}
	return true;
    }

    public Object Current { 
	get { return v; }
    }

    public override string ToString() {
	return "([" + start+ ", " + end + "]/" + step + ", " + v + ")";
    }

    private double start, end, step, v;
    private bool reset;
}

/** 
 * Iterates exhaustively over a list (tuple) of IEnumerables. In every
 * iteration step a new list is created. Attention: this assumes that every
 * element contains at least one element to iterate over.
 */
public class ListIterator : IEnumerator {
    /// Create iterator from a list of IEnumerables.
    public ListIterator( IList list ) {
	if( list.Count == 0 ) {
	    iters = null;
	    return;
	}

	iters = new ArrayList();
	foreach( IEnumerable i in list )
	    iters.Add( i.GetEnumerator() );
	reset = true;
    }

    public bool MoveNext() {
	if( iters == null )
	    return false;

	bool eall = false;
	if( reset ) {
	    reset = false;
	    foreach( IEnumerator i in iters )
		i.MoveNext();
	    eall = true;
	} else {
	    // just iterate until the first element is not finished
	    bool e;
	    foreach( IEnumerator i in iters ) {
		e = i.MoveNext();
		eall = eall || e;
		if( e ) 
		    break;
		else  {
		    i.Reset();
		    i.MoveNext();
		}
	    }
	}

	if( !eall )
	    return false;

	v = new List( iters.Count );
	foreach( IEnumerator i in iters )
	    v.Add( i.Current );
	return true;
    }

    public void Reset() {
	foreach( IEnumerator i in iters )
	    i.Reset();
	reset = true;
    }

    public Object Current { 
	get { return v; }
    }

    public override string ToString() {
	string result="(";
	foreach( IEnumerator i in iters )
	    result += i;
	return result +")";
    }

    /// List of IEnumerator.
    private IList iters;
    private IList v;
    private bool reset;
}


/** 
 * Singleton class for managing exhaustive iteration over all stub property and
 * stub method return values. Stub properties/methods are added lazily to the
 * iteration, i.e. not before they actually appear the first time; this avoids
 * iteration over unused ones.
 */
public class StubIterator {
    /// Initialize iteration.
    public static void start() {
	if( transitionState.Count > 0 || stubInputs.Count > 0 ||
		stubOutputs.Count > 0 || stubCalls.Count > 0 )
	    throw new Exception( "Internal error: attempted to start stub iteration, but internal structures are not empty" );
    }

    /**
     * Iterate to next alternative of input values. Always call this until
     * false is returned.
     * @return true if there are more possibilities, false if last one was
     *         just processed.
     */
    public static bool next() {
	for( int i = transitionState.Count-1; i >= 0; --i ) {
	    if( transitionState[i] is StubOutputState ) {
		stubOutputs.Remove( ( (StubOutputState) transitionState[i]).getKey() );
		transitionState.RemoveAt( i );
	    } 
	    else if( transitionState[i] is StubInputState ) {
		StubInputState sis = (StubInputState) transitionState[i];
		if( sis.MoveNext() )
		    return true;
		stubInputs.Remove( sis.getKey() );
		transitionState.RemoveAt( i );
	    }
	    else {
		StubCallState scs = (StubCallState) transitionState[i];
		if( scs.MoveNext() )
		    return true;
		stubCalls.Remove( scs.getKey() );
		transitionState.RemoveAt( i );
	    } 
	}

	return false;
    }

    /**
     * Return the current value of the stub input field referenced by r. Append
     * a ValueInNode to the current process, if necessary. 
     */
    public static Object stubInput( Ref r ) {
	string stateIdx = r.ToString();
	StubInputState sis = (StubInputState) stubInputs[stateIdx];
	if( sis == null ) {
	    stubInputs[stateIdx] = sis = new StubInputState( r );
	    transitionState.Add( sis );
	}

	sis.execute();
	return sis.Current;
    }

    /**
     * Output value val to stub output field referenced by r. Append a
     * ValueOutNode to the current process, if necessary.
     */
    public static void stubOutput( Ref r, Object val ) {
	StubOutputState sos = (StubOutputState) stubOutputs[r];
	if( sos == null ) {
	    stubOutputs[r] = sos = new StubOutputState( r, val );
	    transitionState.Add( sos );
	} 

	sos.execute();
    }

    /** 
     * Call stub method (given as a Ref) and return the current return value.
     * Append a MethodCallNode to the current process, if necessary. 
     */
    public static Object call( Ref r, List args ) {
	string stateIdx = r + "-" + args;
	StubCallState scs = (StubCallState) stubCalls[stateIdx];
	if( scs == null ) {
	    stubCalls[stateIdx] = scs = new StubCallState( r, args );
	    transitionState.Add( scs );
	}

	scs.execute();
	return scs.Current;
    }

    /**
     * Stores the iteration state of a particular stub input.
     * @see transitionState
     * @see stubInputs
     */
    private class StubInputState {
	Ref r;
	IEnumerator range;
	bool added;
	Node branch;

	/// Create state for given stub input reference.
	public StubInputState( Ref r ) {
	    this.r = r;

	    // get enumerator from stub class field
	    Class stub = r.resolveClass();
	    Class oldCurrClass = Class.currentClass();
	    stub.becomeCurrentClass();
	    this.range = stub.getField( r.getId().next.id ).GetEnumerator();
	    oldCurrClass.becomeCurrentClass();

	    if( !range.MoveNext() )
		throw new Exception( "Stub input property " + r + " has no valid value" );
	    added = false;
	    branch = null;
	}

	/**
	 * Move to next alternative input value.
	 * @return true if there are more values, false if last value was
	 *         already processed
	 */
	public bool MoveNext() {
	    added = false;
	    return range.MoveNext();
	}

	/// Return current input value; undefined if MoveNext() returned false.
	public Object Current {
	    get { return range.Current; }
	}

	/**
	 * Add a ValueInNode to the process if none with the current value
	 * already exists. 
	 */
	public void execute() {
	    if( added )
		return;
	    added = true;

	    if( branch == null )
		branch = Node.currentNode;
	    else
		Node.currentNode = branch;

	    Node.appendCurrent( new ValueInNode( r.getId().id,
			r.getId().next.id, Current ) );
	}

	/**
	 * Return the key value for the stubInputs hashtable.
	 * @see stubInputs
	 */
	public Object getKey() {
	    return r.ToString();
	}

	public override string ToString() {
	    return "StubIn{ property: " + r + ", " +
		(added ? "added" : "not added" ) + 
		", range: " + range + ", branchNode: " + branch + "}";
	}
    }

    /**
     * Stores the iteration state of a particular stub call.
     * @see transitionState
     * @see stubCalls
     */
    private class StubCallState {
	Ref r;
	List args;
	IEnumerator range;
	bool added;
	Node branch;

	/// Create state for given stub method reference and arguments.
	public StubCallState( Ref r, List args ) {
	    this.r = r;
	    this.args = args;

	    // get enumerator from stub class method
	    Class stub = r.resolveClass();
	    Method m = stub.getMethod( r.getId().next.id, args );
	    if( m.returnType() != null ) {
		Class oldCurrClass = Class.currentClass();
		stub.becomeCurrentClass();
		this.range = m.getReturnEnumerator();
		oldCurrClass.becomeCurrentClass();

		if( !range.MoveNext() )
		    throw new Exception( "Stub method " + r + " has no valid return value" );
	    } else
		// void method
		this.range = null;

	    added = false;
	    branch = null;
	}

	/**
	 * Move to next alternative method return value.
	 * @return true if there are more values, false if last value was
	 *         already processed
	 */
	public bool MoveNext() {
	    added = false;
	    if( range != null )
		return range.MoveNext();
	    else
		return false;
	}

	/**
	 * Add a MethodCallNode to the process if none with the current value
	 * already exists. 
	 */
	public void execute() {
	    if( added )
		return;
	    added = true;

	    if( branch == null )
		branch = Node.currentNode;
	    else
		Node.currentNode = branch;

	    if( range != null )
		Node.appendCurrent( new MethodCallNode( r.getId().id,
			    r.getId().next.id, args, Current ) );
	    else
		Node.appendCurrent( new MethodCallNode( r.getId().id,
			    r.getId().next.id, args ) );
	}

	/**
	 * Return the key value for the stubCalls hashtable.
	 * @see stubCalls
	 */
	public Object getKey() {
	    return r.ToString() + "-" + args.ToString();
	}

	/// Return current input value; undefined if MoveNext() returned false.
	public Object Current {
	    get { 
		if( range != null )
		    return range.Current; 
		else
		    return null;
	    }
	}

	public override string ToString() {
	    return "StubCall{ method: " + r + "(" + args + "), " +
		(added ? "added" : "not added" ) + 
		", range: " + range + ", branchNode: " + branch + "}";
	}
    }

    /**
     * Stores the iteration state of a particular stub output.
     * @see transitionState
     * @see stubCalls
     */
    private class StubOutputState {
	Ref r;
	Node ptr;
	Object value;

	/// Create state for given stub output property and value.
	public StubOutputState( Ref r, Object value ) {
	    this.r = r;
	    this.value = value;
	    ptr = null;
	}

	/// Add a ValueOutNode to the process if none already exists. 
	public void execute() {
	    if( ptr == null ) {
		Node.appendCurrent( new ValueOutNode( r.getId().id,
			    r.getId().next.id, value ) );
		ptr = Node.currentNode;
	    } else
		Node.currentNode = ptr;
	}

	/**
	 * Return the key value for the stubOutputs hashtable.
	 * @see stubOutputs
	 */
	public Object getKey() {
	    return r;
	}

	public override string ToString() {
	    return "StubOut{ property: " + r + ", ptr: " + ptr + "}";
	}
    }

    /// Stack of transition states; contains Stub(Input/Output/Call)State.
    private static IList transitionState = new List();

    /// Map transition name (string) to StubInputState for fast lookup.
    private static Hashtable stubInputs = new Hashtable();

    /// Map stub output field (Ref) to StubOutputState for fast lookup.
    private static Hashtable stubOutputs = new Hashtable();

    /// Map transition name (string) to StubCallState for fast lookup.
    private static Hashtable stubCalls = new Hashtable();
}

}

