/**
 * @file process.cs 
 * Classes to represent the generated model as a set of processes.
 *
 * (C) 2004 Martin Pitt <martin@piware.de>
 *
 * This software is distributed under the terms and conditions of the GNU
 * General Public License (see file COPYING)
 */

using System;
using System.IO;
using System.Collections;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;

namespace CCM {

/**
 * Node in the process graph. This encapsulates common behaviour, but only its
 * concrete child classes should (and can) be instantiated.
 */
[Serializable]
public abstract class Node : IEnumerable {
    /// Initialize the node.
    public Node() {
	next = null;
	++numNodes;
    }

    /// Add a new alternative branch n to the node.
    public void addBranch( Node n ) {
	if( next == null ) next = new ArrayList(1);
	next.Add( n );
    }

    /// Return whether this node has any successor nodes.
    public bool hasSuccessors() {
	return next != null;
    }

    /// Generate a CCMB notation of this node and all its successors (i. e.
    /// its transitive alternatives).
    public virtual string ccmb() {
	if( next == null )
	    return "";

	string result = "";
	if( writeDepth++ > 0 )
	    result += " -> ";

	if( next.Count == 1 )
	    result += ( (Node) next[0] ).ccmb();
	else {
	    result += "(";
	    bool first = true;
	    foreach( Node n in next ) {
		result += "\n";
		for( int i = 0; i <= writeDepth; ++i )
		    result += "    ";
		if( first ) {
		    result += "  ";
		    first = false;
		} else
		    result += "| ";
		result += n.ccmb();
	    }
	    result += ")";
	}
	--writeDepth;
	return result;
    }

    /// Generate a Prolog notation of this node and all its successors (i. e.
    /// its transitive alternatives).
    public virtual string prolog() {
	if( next == null )
	    return "";

	string result = "";
	if( writeDepth++ > 0 )
	    result += ", ";
	else
	    result += "[";

	if( next.Count == 1 )
	    result += ( (Node) next[0] ).prolog();
	else {
	    bool first = true;
	    foreach( Node n in next ) {
		result += "\n";
		for( int i = 0; i <= writeDepth; ++i )
		    result += "    ";
		if( first ) {
		    result += "  ";
		    first = false;
		} else
		    result += "+ ";
		result += "[" + n.prolog() + "]";
	    }
	}
	if( --writeDepth == 0 )
	    result += "]";
	
	return result;
    }

    /// Return an enumerator over all child nodes.
    public IEnumerator GetEnumerator() {
	if( next != null )
	    return next.GetEnumerator();
	else
	    return null;
    }

    /// Return a string representation of this Node.
    public override string ToString() {
	string result = "Node";
	if( next != null )
	    result += " (" + next.Count + " children)";
	return result;
    }

    private IList next;
    protected static int writeDepth = 0;

    /// Provide global access to the current node.
    public static Node currentNode = null;

    /// Append given node to the current node and make it the new current node.
    public static void appendCurrent( Node node ) {
	currentNode.addBranch( node );
	currentNode = node;
    }

    public static int numNodes = 0;
}

/**
 * Node that represents a state. CCMB only allows this as the initial 
 * node of a process.
 */
[Serializable]
public class StateNode : Node {
    /// Create state node for given list of Variables.
    public StateNode( IList state ) 
	: base() {
	this.state = state;
	name = null;
    }

    /// Set an user defined name for this state.
    public void setName( string p ) { name = p; }

    /// Return the list of Variables this state represents.
    public IList getState() { return state; }

    /**
     * Returns the name of this state. If an user defined name was set, it is
     * returned; otherwise the name is constructed as a tuple of the values in
     * state.
     * @see setName
     */
    public string stateName() {
	if( name != null )
	    return name;
	return stateStr( state );
    }

    /// Construct a state name tuple from the given object.
    private static string stateStr( Object s ) {
	if( s == null )
	    return "null";
	if( s is bool || s is int || s is double )
	    return s.ToString();
	if( s is Class )
	    return ( (Class) s ).getName();
	if( s is string )
	    return "\"" + s + "\"";

	if( s is IList ) {
	    string result = "[";
	    bool first = true;
	    foreach( Object o in (IList) s ) {
		if( first ) first = false;
		else result += ",";
		result +=  stateStr( o );
	    }
	    return result+"]";
	}

	return s.GetHashCode().ToString();
    }

    /** 
     * Returns the name of this state in Prolog compatible syntax. If an user
     * defined name was set, it is returned; otherwise the name is constructed
     * as a tuple of the values in state.  @see setName
     */
    public string prologStateName() {
	if( name != null )
	    return name.ToLower();
	return "st" + prologStateStr( state );
    }

    /// Construct a Prolog compatible state name tuple from the given object.
    private static string prologStateStr( Object s ) {
	if( s == null )
	    return "null";
	if( s is bool )
	    return ( (bool) s ) ? "true" : "false";
	if( s is int || s is double )
	    return s.ToString();
	if( s is Class )
	    return ( (Class) s ).getName().ToLower();
	if( s is string )
	    return "'" + s + "'";

	if( s is IList ) {
	    string result = "(";
	    bool first = true;
	    foreach( Object o in (IList) s ) {
		if( first ) first = false;
		else result += ",";
		result +=  prologStateStr( o );
	    }
	    return result+")";
	}

	return s.GetHashCode().ToString();
    }

    public override string ToString() {
	return stateName();
    }

    private IList state;
    private string name;
}

/**
 * Node that represents a state reference. CCMB only allows this as a leaf node
 * of a process.
 */
[Serializable]
public class StateRefNode : Node {
    public StateRefNode( string statename ) {
	this.statename = statename;
	--numNodes; // don't count references, but Node already increments
    }

    public override string ToString() {
	return statename;
    }

    public override string ccmb() {
	return statename;
    }

    public override string prolog() {
	StateNode s = (StateNode) Model.currentInstance.getStates()[statename];
	return s.prologStateName();
    }

    private string statename;
}

/**
 * Represents a generic transition node. Must not (and can not) be
 * instantiated directly, use a concrete subclass instead. This class just
 * encapsulates common fields for all transition types.
 */
[Serializable]
public abstract class TransitionNode : Node {
    /**
     * Create an abstract transition that refers to the given stub property and
     * value.
     * @param field Name of the object of this stub (object field)
     * @param prop Stub property/method name
     * @param value Stub property value/method return value
     */
    public TransitionNode( string field, string prop, Object value )
	: base() {
	this.field = field;
	this.prop = prop;
	this.value = value;
    }

    /// Return the stub field name.
    public string getField() { return field; }

    /// Return the stub property/method name.
    public string getProp() { return prop; }

    /// Return the stub property/method return value.
    public Object getValue() { return value; }

    public override string ccmb() {
	return ToString() + base.ccmb();
    }

    public override string prolog() {
	return ToPrologString() + base.prolog();
    }

    public abstract string ToPrologString();

    protected string field, prop;
    protected Object value;
}

/// Represents a stub property reading.
[Serializable]
public class ValueInNode : TransitionNode {
    /**
     * Create an "id = value" input transition.
     * @see TransitionNode.TransitionNode
     */
    public ValueInNode( string field, string prop, Object value )
	: base( field, prop, value ) {}

    public override string ToString() {
	return field + "." + prop + "=" + (value == null ? "null" : value.ToString() );
    }

    public override string ToPrologString() {
	return field.ToLower() + "_" + prop + "=" + (value == null ? "null" : value.ToString().ToLower() );
    }
}

/// Represents a stub property writing.
[Serializable]
public class ValueOutNode : TransitionNode {
    /**
     * Create an "id := value" output transition.
     * @see TransitionNode.TransitionNode
     */
    public ValueOutNode( string field, string prop, Object value )
	: base( field, prop, value ) {}

    public override string ToString() {
	return field + "." + prop + ":=" + (value == null ? "null" : value.ToString() ); 
    }

    public override string ToPrologString() {
	return field.ToLower() + "_" + prop + ":=" + (value == null ? "null" : value.ToString().ToLower() ); 
    }
}

/// Represents a stub method call.
[Serializable]
public class MethodCallNode : TransitionNode {
    /**
     * Create an "id(callargs)" stub transition for void methods. _class and
     * field may be null for component method calls.
     * @see TransitionNode.TransitionNode
     */
    public MethodCallNode( string field, string prop, List callargs )
	: base( field, prop, null ) {
	this.callargs = new List( callargs.Count );
	foreach( Object o in callargs )
	    if( o is Class )
		this.callargs.Add( ( (Class) o ).getName() );
	    else
		this.callargs.Add( o );
	hasReturn = false;
    }

    /// Return the list of call arguments.
    public List getCallArgs() {
	return callargs;
    }

    /**
     * Create a "id(callargs) = value" stub transition for methods returning a
     * value.
     * @see TransitionNode.TransitionNode
     */
    public MethodCallNode( string field, string prop, List callargs, Object value )
	: base( field, prop, value ) {
	this.callargs = new List( callargs.Count );
	foreach( Object o in callargs )
	    if( o is Class )
		this.callargs.Add( ( (Class) o ).getName() );
	    else
		this.callargs.Add( o );
	hasReturn = true;
    }

    public override string ToString() {
	string result = "";
	if( field != null )
	    result += field + ".";
	result += prop + "(" + callargs + ")";
	if( hasReturn )
	    result += "=" + (value == null ? "null" : value.ToString() ); 
	return result;
    }

    public override string ToPrologString() {
	string result;
	if( field != null )
	    result = field.ToLower() + "_" + prop;
	else
	    result = prop.ToLower();
	if( callargs.Count > 0 )
	    result += "(" + callargs + ")";
	if( hasReturn )
	    result += "=" + (value == null ? "null" : value.ToString().ToLower() ); 
	else
	    result = "send(" + result + ")";
	return result;
    }

    private List callargs;
    bool hasReturn;
}

/// Represents a CCMB model (set of processes). 
public class Model {
    /// Initialize an empty model.
    public Model() {
	states = new Hashtable();
	startState = null;
	todo = new Queue();
	comp = null;
	stateAbstraction = null;
	transitionAbstraction = null;
	stateFormat = null;
    }

    /// Deserialize model from the given reader.
    public Model( FileStream reader ) {
	BinaryFormatter fmt = new BinaryFormatter();
	stateFormat = (IList) fmt.Deserialize( reader );
	stateAbstraction = (StateAbstraction) fmt.Deserialize( reader );
	transitionAbstraction = (TransitionAbstraction) fmt.Deserialize( reader );
	string startName = (string) fmt.Deserialize( reader );
	long stateNum = (long) fmt.Deserialize( reader );

	states = new Hashtable();
	StateNode s;
	for( long i = 0; i < stateNum; ++i ) {
	    s = (StateNode) fmt.Deserialize( reader );
	    states.Add( s.stateName(), s );
	}

	startState = (StateNode) states[startName];
	todo = null;
	comp = null;
    }
    
    /// Create a state node from the current state of the component and marks
    /// it as start state.
    public void setStartState() {
	startState = comp.getState();
	startState.setName( comp.getName() );
	states.Add( startState.stateName(), startState );
    }

    /// Return the start state.
    public StateNode getStartState() {
	return startState;
    }

    /// Return the index of the field with given name or -1 if no such field
    /// exists.
    public int fieldIndex( string fieldname ) {
	return stateFormat.IndexOf( fieldname );
    }

    /// Terminate the current process thread by appending a state node and
    /// add it to the todo list (if necessary).
    public void terminateThread() {
	StateNode final;

	final = comp.getState();
	string sname = final.stateName();
	Node.appendCurrent( new StateRefNode( sname ) );

	// only enqueue if sname does not already exist
	if( states.Contains( sname ) )
	    return;
	foreach( StateNode n in todo )
	    if( n.stateName() == sname )
		return;
	todo.Enqueue( final );
    }

    /// Write the generated processes in CCMB notation to the given writer.
    public void writeCCMB( TextWriter writer ) {
	if( comp != null )
	    writer.WriteLine( "State name format: [" + comp.stateFormat() + "]\n" );
	else
	    writer.WriteLine( "State name format: [" + stateFormat + "]\n" );

	writer.WriteLine( "{0} = {1}\n", startState.stateName(), startState.ccmb() );
	foreach( DictionaryEntry i in states )
	    if( i.Value != startState )
		writer.WriteLine( "{0} = {1}\n", i.Key, ( (Node) i.Value).ccmb() );
    }

    /// Write the generated processes in Prolog syntax to the given writer.
    public void writeProlog( TextWriter writer ) {
	// set static variable to allow StateRefNode resolve to a proper
	// StateNode
	currentInstance = this;

	if( comp != null )
	    writer.WriteLine( "% State name format: (" + comp.stateFormat() + ")\n" );

	writer.WriteLine( "processdef( {0}, {1} ).\n", startState.prologStateName(), startState.prolog() );
	StateNode s;
	foreach( DictionaryEntry i in states ) {
	    s = (StateNode) i.Value;
	    if( s != startState )
		writer.WriteLine( "processdef( {0}, {1} ).\n", s.prologStateName(), s.prolog() );
	}
    }

    /// Write the generated processes in binary form to the given stream.
    public void writeBin( Stream writer ) {
	BinaryFormatter fmt = new BinaryFormatter();
	fmt.Serialize( writer, comp.stateFormat() );
	fmt.Serialize( writer, comp.getStateAbstraction() );
	fmt.Serialize( writer, comp.getTransitionAbstraction() );
	fmt.Serialize( writer, startState.stateName() );
	fmt.Serialize( writer, (long) states.Count );
	foreach( DictionaryEntry i in states )
	    fmt.Serialize( writer, i.Value );
    }

    /// Return the next state whose process is not yet calculated and add it to
    /// the set of known states.
    public StateNode nextTodo() {
	StateNode next;

	if( todo.Count == 0 )
	    return null;
	next = (StateNode) todo.Dequeue();
	string nextname = next.stateName();
	if( states.Contains( nextname ) )
	    throw new Exception( "todo item " + nextname + " has already been processed" );
	
	states.Add( nextname, next );
	return next;
    }

    /// Return the state dictionary.
    public Hashtable getStates() {
	return states;
    }

    /**
     * The modelled CCM.Comp class. This field is not serialized! When loading
     * a binary dump, the abstractions will be stored in stateAbstraction and
     * transitionAbstraction.
     */
    public Class comp;

    /// State abstraction in deserialized models.
    public StateAbstraction stateAbstraction;

    /// Transition abstraction in deserialized models.
    public TransitionAbstraction transitionAbstraction;

    /// List of state names in deserialized models.
    public IList stateFormat;

    /// Initial state as set by setStartState().
    private StateNode startState;

    /// Already calculated processes (StateNode.stateName() -> Node).
    private Hashtable states;
      
    /// States whose processes must still be calculated (contains StateNodes).
    private Queue todo;

    /**
     * Used for writeProlog(): StateNodeRef must be able to resolve to a proper
     * StateNode to produce a proper string; this contains a global pointer to
     * the Model instance of which writeProlog() was called.
     */
    public static Model currentInstance;
}

} // closing namespace
